#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <pwd.h>
#include <getopt.h>

#include "dhcp.h"
#include "dhcprogue.h"
#include "socket.h"
#include "mac.h"

static struct option options[] = {
	{ "target-mac",     required_argument, 0, 'm' },
	{ "mac-file",	    required_argument, 0, 'f' },
	{ "spoof-ip",       required_argument, 0, 'i' },
	{ "spoof-dns",      required_argument, 0, 'd' },
	{ "spoof-gateway",  required_argument, 0, 'g' },
	{ "spoof-netmask",  required_argument, 0, 'n' },
	{ "debug",          no_argument,       0, 'D' },
	{ "version",        no_argument,       0, 'v' },
	{ "help"   ,        no_argument,       0, 'h' }
};

void usage(char *prompt) {
	printf("usage: %s [OPTIONS..]\n", prompt);
	printf("\n");
	printf("-m|--target-mac <mac>       only reply for requests from <mac>\n");
	printf("-f|--mac-file <file>        read MAC/IP combinations from <file>\n");
	printf("-i|--spoof-ip <ip>          spoof reply with address <ip>\n");
	printf("-d|--spoof-dns <ip>         spoof reply with dns server <ip>\n");
	printf("-g|--spoof-gateway <ip>     spoof reply with gateway server <ip>\n");
	printf("-n|--spoof-netmask <mask>   spoof reply with netmask <mask>\n");
	printf("-D|--debug                  be extra verbose\n");
	printf("-v|--version                show version and exit\n");
	printf("-h|--help                   show this message\n");
	exit(EXIT_SUCCESS);
}

void version(char *prompt) {
	printf("%s %s\n", prompt, VERSION);
	exit(EXIT_SUCCESS);
}

void fatal(char *format, ...) {
	
	char buffer[1024];
	va_list args;
	
	memset(&buffer, 0, sizeof(buffer));
	
	va_start(args, format);
	vsnprintf(buffer, sizeof(buffer) - 1, format, args);
	va_end(args);

	printf("fatal error: %s\n", buffer);
	terminate(EXIT_FAILURE);
}

void msg(char *format, ...) {

	char buffer[1024];
	va_list args;
	
	memset(&buffer, 0, sizeof(buffer));
	
	va_start(args, format);
	vsnprintf(buffer, sizeof(buffer) - 1, format, args);
	va_end(args);

	printf("%s[%d]: %s\n", prog, getpid(), buffer);
	fflush(stdout);
}

void debug(char *format, ...) {

	if(debugmode) {

		char buffer[1024];
		va_list args;
	
		memset(&buffer, 0, sizeof(buffer));
	
		va_start(args, format);
		vsnprintf(buffer, sizeof(buffer) - 1, format, args);
		va_end(args);

		printf("%s[%d]: %s\n", prog, getpid(), buffer);
		fflush(stdout);

	}
}

void terminate(int err) {

	if(listener)
		socket_deinit(listener);
	
	exit(err);
}

void drop_privileges() {

	struct passwd *pw;

	pw = getpwnam("nobody");
	if(!pw)
		fatal("getpwnam() failed: %s", strerror(errno));
	
	if(setgid(pw->pw_gid) != 0)
		fatal("setgid() failed: %s", strerror(errno));
	if(setuid(pw->pw_uid) != 0)
		fatal("setuid() failed; %s", strerror(errno));
	else
		msg("dropped privileges successfully to %d:%d (nobody)", pw->pw_uid, pw->pw_gid);
}

int main(int argc, char **argv) {

	int index, c;

	prog = argv[0];
	target_mac = 0;
	spoof_ip = 0;
	spoof_dns = 0;
	spoof_gateway = 0;
	spoof_netmask = "255.255.255.0";
	debugmode = FALSE;

	while(1) {
		c = getopt_long(argc, argv, "m:f:i:d:g:Dvh", options, &index);

		if(c == -1)
			break;
		
		switch(c) {
			case 'm':
				target_mac = optarg;
				break;
			case 'f':
				mac_file = optarg;
				break;
			case 'i':
				spoof_ip = optarg;
				break;
			case 'd':
				spoof_dns = optarg;
				break;
			case 'g':
				spoof_gateway = optarg;
				break;
			case 'n':
				spoof_netmask = optarg;
				break;
			case 'D':
				debugmode = TRUE;
				break;
			case 'v':
				version(prog);
				break;
			case 'h':
				usage(prog);
				break;
		}
	}

	if(!mac_file && !spoof_ip)
		fatal("you must specify at least the spoofed IP address (-i) or the MAC/IP combinations file (-f)");
	else if(mac_file)
		mac_file_init(mac_file);
	if(!spoof_dns)
		fatal("you must specify at least the spoofed DNS server (-d)");
	if(!spoof_gateway)
		fatal("you must specify at least the spoofed IP gateway (-g)");

	msg("initializing (MAC=%s, MACFILE=%s, IP=%s, DNS=%s, GATEWAY=%s, NETMASK=%s)", target_mac, mac_file, spoof_ip, spoof_dns, spoof_gateway, spoof_netmask);
	
	if(geteuid() != 0)
		fatal("you must be superuser to run %s", prog);
	
	listener = socket_init(PORT);
	drop_privileges();
	msg("");
	dhcp_loop_packets(listener);

	terminate(EXIT_SUCCESS);
	return -1;
}
